import pytest
import pandas as pd
import numpy as np
from pydisagg.ihme.validator import (
    validate_columns,
    validate_index,
    validate_nonan,
    validate_positive,
    validate_interval,
    validate_noindexdiff,
)


@pytest.fixture
def data():
    np.random.seed(123)
    return pd.DataFrame(
        dict(
            uid=range(10),
            sex_id=[1] * 5 + [2] * 5,
            location_id=[1, 2] * 5,
            year_id=[2010] * 10,
            age_start=[0, 5, 10, 17, 20] * 2,
            age_end=[12, 10, 22, 21, 25] * 2,
            val=5.0,
            val_sd=1.0,
        )
    )


@pytest.fixture
def pattern():
    np.random.seed(123)
    return pd.DataFrame(
        dict(
            sex_id=[1] * 5 + [2] * 5,
            age_start=[0, 5, 10, 15, 20] * 2,
            age_end=[5, 10, 15, 20, 25] * 2,
            age_group_id=list(range(5)) * 2,
            draw_0=np.random.rand(10),
            draw_1=np.random.rand(10),
            draw_2=np.random.rand(10),
            year_id=[2010] * 10,
            location_id=[1, 2] * 5,
        )
    )


@pytest.fixture
def population():
    np.random.seed(123)
    sex_id = pd.DataFrame(dict(sex_id=[1, 2]))
    year_id = pd.DataFrame(dict(year_id=[2010]))
    location_id = pd.DataFrame(dict(location_id=[1, 2]))
    age_group_id = pd.DataFrame(dict(age_group_id=range(5)))

    population = (
        sex_id.merge(location_id, how="cross")
        .merge(age_group_id, how="cross")
        .merge(year_id, how="cross")
    )
    population["population"] = 1000
    return population


def test_validate_columns_missing(population):
    with pytest.raises(KeyError):
        validate_columns(
            population.drop(columns=["population"]),
            ["sex_id", "location_id", "age_group_id", "year_id", "population"],
            "population",
        )


def test_validate_index_missing(population):
    with pytest.raises(ValueError):
        validate_index(
            pd.concat([population, population]),
            ["sex_id", "location_id", "age_group_id", "year_id"],
            "population",
        )


def test_validate_nonan(population):
    with pytest.raises(ValueError):
        validate_nonan(population.assign(population=np.nan), "population")


def test_validate_positive_strict(population):
    with pytest.raises(ValueError):
        validate_positive(
            population.assign(population=0),
            ["population"],
            "population",
            strict=True,
        )


def test_validate_positive_not_strict(population):
    with pytest.raises(ValueError):
        validate_positive(
            population.assign(population=-1),
            ["population"],
            "population",
            strict=False,
        )


def test_validate_interval_lower_equal_upper(data):
    with pytest.raises(ValueError):
        validate_interval(
            data.assign(age_end=data["age_start"]),
            "age_start",
            "age_end",
            ["uid"],
            "data",
        )


def test_validate_interval_lower_greater_than_upper(data):
    with pytest.raises(ValueError):
        validate_interval(
            data.assign(age_end=0), "age_start", "age_end", ["uid"], "data"
        )


def test_validate_interval_positive(data):
    validate_interval(data, "age_start", "age_end", ["uid"], "data")


def test_validate_positive_no_error(population):
    validate_positive(population, ["population"], "population", strict=True)
    validate_positive(population, ["population"], "population", strict=False)


@pytest.fixture
def merged_data_pattern(data, pattern):
    return pd.merge(
        pattern,
        data,
        on=["sex_id", "age_start", "age_end", "location_id"],
        how="outer",
    )


def test_validate_noindexdiff_merged_positive(merged_data_pattern, population):
    # Positive test case: no index difference
    validate_noindexdiff(
        population,
        merged_data_pattern,
        ["sex_id", "location_id"],
        "merged_data_pattern",
    )


def test_validate_noindexdiff_merged_negative(data, pattern):
    # Negative test case: index difference exists
    data_with_pattern = data.merge(
        pattern,
        on=["sex_id", "age_start", "age_end", "location_id", "year_id"],
        how="left",
    )
    # Introduce an index difference
    data_with_pattern = data_with_pattern.drop(index=0)
    with pytest.raises(ValueError):
        validate_noindexdiff(
            data,
            data_with_pattern,
            data.columns.tolist(),
            "merged_data_pattern",
        )



# TODO Fix the disaggregate.py file and then fill this in to make this work
# TODO Also add dataframe splitting stuff
import numpy as np
import pytest
from numpy.testing import assert_approx_equal

from pydisagg import models
from pydisagg.disaggregate import split_datapoint

model_list = [
    models.RateMultiplicativeModel(),
    models.LogOdds_model(),
    models.LMO_model(5),
]


@pytest.mark.parametrize("model", model_list)
def test_count_model_consistency(model):
    populations = np.array([2, 5])
    measured_total = 4.8
    measurement_SE = 1
    rate_pattern = np.array([0.2, 0.4])

    result, SE = split_datapoint(
        measured_total,
        populations,
        rate_pattern,
        measurement_SE,
        model,
        output_type="total",
    )
    assert_approx_equal(measured_total, np.sum(result))
    assert_approx_equal(measurement_SE, np.sum(SE))


@pytest.mark.parametrize("model", model_list)
def test_rate_model_consistency(model):
    populations = np.array([2, 5])
    measured_total = 4.8
    measurement_SE = 1
    rate_pattern = np.array([0.2, 0.4])

    result, SE = split_datapoint(
        measured_total,
        populations,
        rate_pattern,
        measurement_SE,
        model,
        output_type="rate",
    )
    assert_approx_equal(measured_total, np.sum(result * populations))



"""
Module containing classes of ParameterTransformations
"""

from typing import Union

import numpy as np
from numpy.typing import NDArray
from scipy.optimize import root_scalar

from pydisagg.ParameterTransformation import ParameterTransformation

float_or_array = Union[float, NDArray]


class LogTransformation(ParameterTransformation):
    """
    Logarithmic transformation of parameter
    Since we fit models additively in beta the log makes it multiplicative
    """

    def __call__(self, x: float_or_array):
        return np.log(x)

    def inverse(self, z: float_or_array):
        return np.exp(z)

    def diff(self, x: float_or_array):
        return 1 / x


class LogModifiedOddsTransformation(ParameterTransformation):
    """
    Log Modified odds transformation
    T(x)=log(x/(1-x**a))
    """

    def __init__(self, a: float):
        self.a = a

    def __call__(self, x: float_or_array):
        """
        Calls transformation function
        """
        return np.log(x / (1 - (x**self.a)))

    def _inverse_single(self, z: float):
        def root_func(x):
            return np.exp(z) * (1 - x**self.a) - x

        return root_scalar(root_func, bracket=[0, 1], method="toms748").root

    def inverse(self, z: float_or_array):
        return np.vectorize(self._inverse_single)(z)

    def diff(self, x: float_or_array):
        numerator = (self.a - 1) * (x**self.a) + 1
        denominator = ((x**self.a) - 1) * x
        return -1 * numerator / denominator


class LogOddsTransformation(ParameterTransformation):
    """
    Log-odds transformation
    T(x)=log(x/(1-x))
    """

    def __call__(self, x: float_or_array):
        """
        Calls transformation function
        """
        return np.log(x / (1 - x))

    def inverse(self, z: float_or_array):
        expz = np.exp(z)
        return expz / (1 + expz)

    def diff(self, x: float_or_array):
        return 1 / (x - x**2)



import pytest
import pandas as pd
import numpy as np
from pydantic import ValidationError
from pydisagg.ihme.splitter import (
    AgeSplitter,
    DataConfig,
    PatternConfig,
    PopulationConfig,
)


@pytest.fixture
def data():
    np.random.seed(123)
    return pd.DataFrame(
        dict(
            uid=range(10),
            sex_id=[1] * 5 + [2] * 5,
            location_id=[1, 2] * 5,
            year_id=[2010] * 10,
            age_start=[0, 5, 10, 17, 20] * 2,
            age_end=[12, 10, 22, 21, 25] * 2,
            val=5.0,
            val_sd=1.0,
        )
    )


@pytest.fixture
def pattern():
    np.random.seed(123)
    pattern_df1 = pd.DataFrame(
        dict(
            sex_id=[1] * 5 + [2] * 5,
            age_start=[0, 5, 10, 15, 20] * 2,
            age_end=[5, 10, 15, 20, 25] * 2,
            age_group_id=list(range(5)) * 2,
            draw_0=np.random.rand(10),
            draw_1=np.random.rand(10),
            draw_2=np.random.rand(10),
            year_id=[2010] * 10,
            location_id=[1]*10,
        )
    )
    pattern_df2 = pattern_df1.copy()
    pattern_df2['location_id']=2
    return pd.concat([pattern_df1,pattern_df2]).reset_index(drop=True)

@pytest.fixture
def population():
    np.random.seed(123)
    sex_id = pd.DataFrame(dict(sex_id=[1, 2]))
    year_id = pd.DataFrame(dict(year_id=[2010]))
    location_id = pd.DataFrame(dict(location_id=[1, 2]))
    age_group_id = pd.DataFrame(dict(age_group_id=range(5)))

    population = (
        sex_id.merge(location_id, how="cross")
        .merge(age_group_id, how="cross")
        .merge(year_id, how="cross")
    )
    population["population"] = 1000
    return population


def test_model_post_init():
    data_config = DataConfig(
        index=["index1"],
        age_lwr="age_lwr1",
        age_upr="age_upr1",
        val="val1",
        val_sd="val_sd1",
    )
    pattern_config = PatternConfig(
        by=["by1"],
        age_key="age_key1",
        age_lwr="age_lwr2",
        age_upr="age_upr2",
        draws=["draw1"],
        draw_mean="draw_mean1",
        draw_var="draw_var1",
        prefix="prefix1",
    )
    population_config = PopulationConfig(
        index=["index2"], val="val2", prefix="prefix2"
    )

    # Test when pattern.by is not a subset of data.index
    with pytest.raises(ValidationError):
        AgeSplitter(
            data=data_config,
            pattern=pattern_config,
            population=population_config,
        )

    # Test when pattern.age_key is in data.index
    data_config.index.append("age_key1")
    pattern_config.by.append("index1")
    with pytest.raises(ValidationError):
        AgeSplitter(
            data=data_config,
            pattern=pattern_config,
            population=population_config,
        )

    # Test when pattern.age_key is not in population.index
    data_config.index.remove("age_key1")
    population_config.index.remove("index2")
    with pytest.raises(ValidationError):
        AgeSplitter(
            data=data_config,
            pattern=pattern_config,
            population=population_config,
        )

    # Test when population.index is not a subset of data.index + pattern.index
    population_config.index.append("index3")
    with pytest.raises(ValidationError):
        AgeSplitter(
            data=data_config,
            pattern=pattern_config,
            population=population_config,
        )


@pytest.mark.skip(reason="not implemented yet")
def test_parse_data(data):
    data_config = DataConfig(
        index=["unique_id", "location_id", "year_id", "sex_id"],
        age_lwr="age_start",
        age_upr="age_end",
        val="mean",
        val_sd="SE",
    )
    splitter = AgeSplitter(data=data_config, pattern=None, population=None)
    parsed_data = splitter.parse_data(data)
    assert isinstance(parsed_data, pd.DataFrame)
    assert not parsed_data.empty
    assert set(data_config.index).issubset(parsed_data.columns)
    assert data_config.age_lwr in parsed_data.columns
    assert data_config.age_upr in parsed_data.columns
    assert data_config.val in parsed_data.columns
    assert data_config.val_sd in parsed_data.columns


@pytest.mark.skip(reason="not implemented yet")
def test_parse_pattern(pattern):
    splitter = AgeSplitter()
    # Add assertions for each test case


@pytest.mark.skip(reason="not implemented yet")
def test_merge_with_pattern(data, pattern):
    splitter = AgeSplitter()
    # Add assertions for each test case


@pytest.mark.skip(reason="not implemented yet")
def test_parse_population(population):
    splitter = AgeSplitter()
    # Add assertions for each test case


@pytest.mark.skip(reason="not implemented yet")
def test_merge_with_population(data, population):
    splitter = AgeSplitter()
    # Add assertions for each test case


@pytest.mark.skip(reason="not implemented yet")
def test_align_pattern_and_population(data):
    splitter = AgeSplitter()
    # Add assertions for each test case


@pytest.mark.skip(reason="not implemented yet")
def test_split(data, pattern, population):
    splitter = AgeSplitter()
    # Add assertions for each test case



from typing import Any

import numpy as np
from pandas import DataFrame
from pydantic import BaseModel

from pydisagg.disaggregate import split_datapoint
from pydisagg.ihme.validator import (
    validate_columns,
    validate_index,
    validate_interval,
    validate_noindexdiff,
    validate_nonan,
    validate_positive,
)
from pydisagg.models import LogOdds_model, RateMultiplicativeModel


class DataConfig(BaseModel):
    index: list[str]
    age_lwr: str
    age_upr: str
    val: str
    val_sd: str

    @property
    def columns(self) -> list[str]:
        return self.index + [self.age_lwr, self.age_upr, self.val, self.val_sd]


class PatternConfig(BaseModel):
    by: list[str]
    age_key: str
    age_lwr: str
    age_upr: str

    draws: list[str] = []
    val: str = "val"
    val_sd: str = "val_sd"
    prefix: str = "pat_"

    @property
    def index(self) -> list[str]:
        return self.by + [self.age_key]

    @property
    def columns(self) -> list[str]:
        return self.index + [
            self.age_lwr,
            self.age_upr,
            self.val,
            self.val_sd,
        ]

    @property
    def val_fields(self) -> list[str]:
        return [
            "age_lwr",
            "age_upr",
            "val",
            "val_sd",
        ]

    def apply_prefix(self) -> dict[str, str]:
        rename_map = {}
        for field in self.val_fields:
            new_field_val = self.prefix + (field_val := getattr(self, field))
            rename_map[field_val] = new_field_val
            setattr(self, field, new_field_val)
        return rename_map

    def remove_prefix(self) -> None:
        for field in self.val_fields:
            setattr(self, field, getattr(self, field).removeprefix(self.prefix))


class PopulationConfig(BaseModel):
    index: list[str]
    val: str

    prefix: str = "pop_"

    @property
    def columns(self) -> list[str]:
        return self.index + [self.val]

    @property
    def val_fields(self) -> list[str]:
        return ["val"]

    def apply_prefix(self) -> dict[str, str]:
        rename_map = {}
        for field in self.val_fields:
            new_field_val = self.prefix + (field_val := getattr(self, field))
            rename_map[field_val] = new_field_val
            setattr(self, field, new_field_val)
        return rename_map

    def remove_prefix(self) -> None:
        for field in self.val_fields:
            setattr(self, field, getattr(self, field).removeprefix(self.prefix))


class AgeSplitter(BaseModel):
    data: DataConfig
    pattern: PatternConfig
    population: PopulationConfig

    def model_post_init(self, __context: Any) -> None:
        """Extra validation of all the index."""
        if not set(self.pattern.by).issubset(self.data.index):
            raise ValueError("pattern.by must be a subset of data.index")
        if self.pattern.age_key in self.data.index:
            raise ValueError("pattern.age_key must not be in data.index")
        if self.pattern.age_key not in self.population.index:
            raise ValueError("pattern.age_key must be in population.index")
        if not set(self.population.index).issubset(
            self.data.index + self.pattern.index
        ):
            raise ValueError(
                "population.index must be a subset of data.index + pattern.index"
            )

    def parse_data(self, data: DataFrame) -> DataFrame:
        name = "data"
        validate_columns(data, self.data.columns, name)

        data = data[self.data.columns].copy()

        validate_index(data, self.data.index, name)
        validate_nonan(data, name)
        validate_positive(data, [self.data.val_sd], name)
        validate_interval(
            data, self.data.age_lwr, self.data.age_upr, self.data.index, name
        )
        return data

    def parse_pattern(self, data: DataFrame, pattern: DataFrame) -> DataFrame:
        name = "pattern"

        if not all(
            col in pattern.columns
            for col in [self.pattern.val, self.pattern.val_sd]
        ):
            if not self.pattern.draws:
                raise ValueError(
                    "Must provide draws for pattern if pattern.val and "
                    "pattern.val_sd are not available."
                )

            validate_columns(pattern, self.pattern.draws, name)
            pattern[self.pattern.val] = pattern[self.pattern.draws].mean(axis=1)
            pattern[self.pattern.val_sd] = pattern[self.pattern.draws].std(
                axis=1
            )

        validate_columns(pattern, self.pattern.columns, name)
        pattern = pattern[self.pattern.columns].copy()

        validate_index(pattern, self.pattern.index, name)
        validate_nonan(pattern, name)
        validate_positive(pattern, [self.pattern.val_sd], name)
        validate_interval(
            pattern,
            self.pattern.age_lwr,
            self.pattern.age_upr,
            self.pattern.index,
            name,
        )

        rename_map = self.pattern.apply_prefix()
        pattern.rename(columns=rename_map, inplace=True)

        data_with_pattern = self._merge_with_pattern(data, pattern)

        validate_noindexdiff(data, data_with_pattern, self.data.index, name)
        # TODO: add validation checks for incomplete age pattern
        # * pattern age intervals do not overlap
        # * smallest pattern interval doesn't cover the left end point of data
        # * largest pattern interval doesn't cover the right end point of data
        # How to vectorize this action...
        return data_with_pattern

    def _merge_with_pattern(
        self, data: DataFrame, pattern: DataFrame
    ) -> DataFrame:
        data_with_pattern = (
            data.merge(pattern, on=self.pattern.by, how="left")
            .query(
                f"({self.pattern.age_lwr} >= {self.data.age_lwr} and"
                f" {self.pattern.age_lwr} < {self.data.age_upr}) or"
                f"({self.pattern.age_upr} > {self.data.age_lwr} and"
                f" {self.pattern.age_upr} <= {self.data.age_upr})"
            )
            .dropna()
        )
        return data_with_pattern

    def parse_population(
        self, data: DataFrame, population: DataFrame
    ) -> DataFrame:
        name = "population"
        validate_columns(population, self.population.columns, name)

        population = population[self.population.columns].copy()

        validate_index(population, self.population.index, name)
        validate_nonan(population, name)

        rename_map = self.population.apply_prefix()
        population.rename(columns=rename_map, inplace=True)

        data_with_population = self._merge_with_population(data, population)

        validate_noindexdiff(
            data,
            data_with_population,
            self.data.index + [self.pattern.age_key],
            name,
        )
        return data_with_population

    def _merge_with_population(
        self, data: DataFrame, population: DataFrame
    ) -> DataFrame:
        data_with_population = data.merge(
            population,
            on=self.population.index,
            how="left",
        ).dropna()
        return data_with_population

    def _align_pattern_and_population(self, data: DataFrame) -> DataFrame:
        data = data.sort_values(
            self.data.index + [self.pattern.age_lwr, self.pattern.age_upr],
            ignore_index=True,
        )

        data[self.pattern.val + "_aligned"] = data[self.pattern.val]
        data[self.pattern.val_sd + "_aligned"] = data[self.pattern.val_sd]
        data[self.population.val + "_aligned"] = data[self.population.val]

        data_group = data.groupby(self.data.index, sort=False)
        for key, data_sub in data_group:
            index_first, index_last = data_group.groups[key][[0, -1]]
            data_first, data_last = data.loc[index_first], data.loc[index_last]

            # TODO: currently we do constant interpolation for pattern
            data.loc[index_first, self.pattern.val + "_aligned"] = data_first[
                self.pattern.val
            ]
            data.loc[index_first, self.pattern.val_sd + "_aligned"] = (
                data_first[self.pattern.val_sd]
            )
            data.loc[index_last, self.pattern.val + "_aligned"] = data_last[
                self.pattern.val
            ]
            data.loc[index_last, self.pattern.val_sd + "_aligned"] = data_last[
                self.pattern.val_sd
            ]

            # align population
            # TODO: this is naive implementation compute the proprotion of population
            # within the given age interval
            data.loc[index_first, self.population.val + "_aligned"] = (
                data_first[self.population.val]
                / (
                    data_first[self.pattern.age_upr]
                    - data_first[self.pattern.age_lwr]
                )
                * (
                    data_first[self.pattern.age_upr]
                    - data_first[self.data.age_lwr]
                )
            )
            data.loc[index_last, self.population.val + "_aligned"] = (
                data_last[self.population.val]
                / (
                    data_last[self.pattern.age_upr]
                    - data_last[self.pattern.age_lwr]
                )
                * (
                    data_last[self.data.age_upr]
                    - data_last[self.pattern.age_lwr]
                )
            )

        return data

    def split(
        self,
        data: DataFrame,
        pattern: DataFrame,
        population: DataFrame,
        model: str = "rate",
        output_type: str = "rate",
    ) -> DataFrame:
        """
        Splits the data based on the given pattern and population. The split results are added to the data as new columns.

        Parameters
        ----------
        data : DataFrame
            The data to be split.
        pattern : DataFrame
            The pattern to be used for splitting the data.
        population : DataFrame
            The population to be used for splitting the data.
        model : str, optional
            The model to be used for splitting the data, by default "rate". Can be "rate" or "logodds".
        output_type : str, optional
            The type of output to be returned, by default "rate".

        Returns
        -------
        DataFrame
            The two main output columns are: 'split_result' and 'split_result_se'.
            There are additional intermediate columns for sanity checks and calculations (have a prefix of pat_ or pop_, and a suffix of _aligned).

        """
        model_mapping = {
            "rate": RateMultiplicativeModel(),
            "logodds": LogOdds_model(),
        }

        if model not in model_mapping:
            raise ValueError(
                f"Invalid model: {model}. Expected one of: {list(model_mapping.keys())}"
            )

        model_instance = model_mapping[model]

        data = self.parse_data(data)
        data = self.parse_pattern(data, pattern)
        data = self.parse_population(data, population)

        data = self._align_pattern_and_population(data)

        # where split happens
        data["split_result"], data["split_result_se"] = np.nan, np.nan
        data_group = data.groupby(self.data.index)
        for key, data_sub in data_group:
            split_result, SE = split_datapoint(
                observed_total=data_sub[self.data.val].iloc[0],
                bucket_populations=data_sub[
                    self.population.val + "_aligned"
                ].to_numpy(),
                rate_pattern=data_sub[self.pattern.val + "_aligned"].to_numpy(),
                model=model_instance,
                output_type=output_type,
                normalize_pop_for_average_type_obs=True,
                observed_total_se=data_sub[self.data.val_sd].iloc[0],
                pattern_covariance=np.diag(
                    data_sub[self.pattern.val_sd + "_aligned"].to_numpy() ** 2
                ),
            )
            index = data_group.groups[key]
            data.loc[index, "split_result"] = split_result
            data.loc[index, "split_result_se"] = SE

        self.pattern.remove_prefix()
        self.population.remove_prefix()

        return data



import pandas as pd
from pandas import DataFrame


def validate_columns(df: DataFrame, columns: list[str], name: str) -> None:
    missing = [col for col in columns if col not in df.columns]
    if missing:
        raise KeyError(f"{name} has missing columns: {missing}")


def validate_index(df: DataFrame, index: list[str], name: str) -> None:
    duplicated_index = pd.MultiIndex.from_frame(
        df[df[index].duplicated()][index]
    ).to_list()
    if duplicated_index:
        raise ValueError(f"{name} has duplicated index: {duplicated_index}")


def validate_nonan(df: DataFrame, name: str) -> None:
    nan_columns = df.columns[df.isna().any(axis=0)].to_list()
    if nan_columns:
        raise ValueError(f"{name} has NaN values in columns: {nan_columns}")


def validate_positive(
    df: DataFrame, columns: list[str], name: str, strict: bool = True
) -> None:
    op = "<=" if strict else "<"
    negative = [col for col in columns if df.eval(f"{col} {op} 0").any()]
    if negative:
        raise ValueError(f"{name} has negative values in: {negative}")


def validate_interval(
    df: DataFrame, lwr: str, upr: str, index: list[str], name: str
) -> None:
    invalid_index = pd.MultiIndex.from_frame(
        df.query(f"{lwr} >= {upr}")[index]
    ).to_list()
    if invalid_index:
        raise ValueError(
            f"{name} has invalid interval with index: {invalid_index}"
        )


def validate_noindexdiff(
    df_ref: DataFrame, df: DataFrame, index: list[str], name: str
) -> None:
    index_ref = pd.MultiIndex.from_frame(df_ref[index])
    index = pd.MultiIndex.from_frame(df[index])
    missing_index = index_ref.difference(index).to_list()

    if missing_index:
        raise ValueError(f"Missing {name} info for index: {missing_index}")



"""
Module containing abstract ParameterTransformation class
"""

from abc import ABC, abstractmethod
from typing import Union

from numpy.typing import NDArray

float_or_array = Union[float, NDArray]


class ParameterTransformation(ABC):
    """
    Transformation that contains function, first order and inverse information
    """

    @abstractmethod
    def __call__(self, x: float_or_array):
        """
        Calls transformation function
        """

    @abstractmethod
    def inverse(self, z: float_or_array):
        """
        Calls inverse of transformation function
        """

    @abstractmethod
    def diff(self, x: float_or_array):
        """
        Calls derivative of transformation
        """
